# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Iterator

from troubadix.helper import is_ignore_file
from troubadix.helper.patterns import ScriptTag, get_script_tag_pattern
from troubadix.plugin import FileContentPlugin, LinterError, LinterResult

# We don't want to touch the metadata of this older VTs...
_IGNORE_FILES = [
    "nmap_nse/",
]


class CheckMissingTagSolution(FileContentPlugin):
    name = "check_missing_tag_solution"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        """This script checks if the VT has a solution_type script tag:
        script_tag(name:"solution_type", value:"");

        but is missing a solution tag within the description block like:
        script_tag(name:"solution", value:"");

        This excludes files from the (sub)dir "nmap_nse/" and deprecated
        vts.
        """

        if nasl_file.suffix == ".inc":
            return

        if is_ignore_file(nasl_file, _IGNORE_FILES):
            return

        # Not all VTs have/require a solution_type text
        if "solution_type" not in file_content:
            return

        # Avoid unnecessary message against deprecated VTs.
        deprecated_pattern = get_script_tag_pattern(ScriptTag.DEPRECATED)
        deprecated_match = deprecated_pattern.search(string=file_content)

        if deprecated_match and deprecated_match.group("value"):
            return

        solution_type_pattern = get_script_tag_pattern(ScriptTag.SOLUTION_TYPE)
        solution_type_match = solution_type_pattern.search(string=file_content)
        if not solution_type_match:
            return

        solution_pattern = get_script_tag_pattern(ScriptTag.SOLUTION)
        solution_match = solution_pattern.search(string=file_content)

        if not solution_match or solution_match.group(0) is None:
            yield LinterError(
                "'solution_type' script_tag but no 'solution' script_tag "
                "found in the description block.",
                file=nasl_file,
                plugin=self.name,
            )



# Copyright (C) 2021 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from troubadix.helper import CURRENT_ENCODING
from troubadix.plugin import LinterError
from troubadix.plugins.badwords import CheckBadwords

from . import PluginTestCase

root = Path(__file__).parent / "test_files"


class TestBadwords(PluginTestCase):
    def test_files(self):
        nasl_file = root / "fail_badwords.nasl"
        context = self.create_file_plugin_context(
            nasl_file=nasl_file,
            lines=nasl_file.read_text(encoding=CURRENT_ENCODING).splitlines(),
        )
        plugin = CheckBadwords(context)

        results = list(plugin.run())

        self.assertEqual(len(results), 3)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "Badword in line     1: openvas is a bad word",
            results[0].message,
        )
        self.assertEqual(
            "Badword in line     6: # OpenVAS Vulnerability Test",
            results[1].message,
        )
        self.assertEqual(
            "Badword in line    10: OpenVAS is a scanner",
            results[2].message,
        )

    def test_combined(self):
        path = Path("some/find_service3.nasl")
        content = "  OpenVAS-8 and probably prior\nOpenVAS-9"

        fake_context = self.create_file_plugin_context(
            nasl_file=path, lines=content.splitlines()
        )

        plugin = CheckBadwords(fake_context)
        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_exception_ok(self):
        path = Path("some/include.inc")
        content = '# HostDetails/NVT                 => "1.2.3.4"'

        fake_context = self.create_file_plugin_context(
            nasl_file=path, lines=content.splitlines()
        )

        plugin = CheckBadwords(fake_context)
        results = list(plugin.run())

        self.assertEqual(len(results), 0)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from troubadix.plugin import LinterError
from troubadix.plugins.grammar import CheckGrammar

from . import PluginTestCase


class CheckNewlinesTestCase(PluginTestCase):
    def test_ok(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"vuldetect", value:"Sends multiple HTTP GET '
            'requests and checks the responses.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_grammar(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
            "# is prone to a security bypass vulnerabilities\n"
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem: "
            "# is prone to a security bypass vulnerabilities",
            results[0].message,
        )

    def test_grammar2(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
            "# is prone to a security bypass vulnerabilities\n"
            "# refer the Reference\n"
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 2)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem: "
            "# is prone to a security bypass vulnerabilities",
            results[0].message,
        )

        self.assertIsInstance(results[1], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem: "
            "# refer the Reference",
            results[1].message,
        )

    def test_grammar3(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
            '  script_tag(name:"summary", value:"Adobe Digital Edition is '
            'prone a to denial of service (DoS) vulnerability.");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem:   "
            'script_tag(name:"summary", value:"Adobe Digital Edition is prone '
            'a to denial of service (DoS) vulnerability.");',
            results[0].message,
        )

    def test_grammar4(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
            '  script_tag(name:"summary", value:"Splunk Enterprise is prone an '
            'open redirect vulnerability.");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem:   "
            'script_tag(name:"summary", value:"Splunk Enterprise is prone an '
            'open redirect vulnerability.");',
            results[0].message,
        )

    def test_grammar5(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
            '  script_tag(name:"vuldetect", value:"Sends multiple HTTP GET '
            'request and checks the responses.");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem:   "
            'script_tag(name:"vuldetect", value:"Sends multiple HTTP GET '
            'request and checks the responses.");',
            results[0].message,
        )

    def test_grammar6(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar is prone to multiple '
            'unknown vulnerability.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )

        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            "VT/Include has the following grammar problem:   "
            'script_tag(name:"summary", value:"Foo Bar is prone to multiple '
            'unknown vulnerability.");',
            results[0].message,
        )

    def test_grammar_fp(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.'
            ' a multiple keyboard .");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_grammar_fp1(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar is prone to '
            ' multiple cross-site request forgery (CSRF) vulnerabilities.");\n'
            '  script_tag(name:"insight", value:"A Cross Site Request '
            ' Forgery flaw exists.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=nasl_file, file_content=content
        )
        plugin = CheckGrammar(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path

from troubadix.plugin import LinterError, LinterWarning
from troubadix.plugins.illegal_characters import CheckIllegalCharacters

from . import PluginTestCase


class CheckIllegalCharactersTestCase(PluginTestCase):
    def test_ok(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo Bar.");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckIllegalCharacters(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_exclude_inc_file(self):
        path = Path("some/file.inc")
        fake_context = self.create_file_plugin_context(nasl_file=path)
        plugin = CheckIllegalCharacters(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_illegal_chars_in_various_tags(self):
        tags = [
            "summary",
            "impact",
            "affected",
            "insight",
            "vuldetect",
            "solution",
        ]
        path = Path("tests/file.nasl")
        for tag in tags:
            content = (
                '  script_tag(name:"cvss_base", value:"4.0");\n'
                f'  script_tag(name:"{tag}", value:"Foo|Bar;Baz=Bad.");\n'
                '  script_tag(name:"solution_type", value:"VendorFix");\n'
                '  script_tag(name:"solution", value:"meh");\n'
            )

            fake_context = self.create_file_plugin_context(
                nasl_file=path, file_content=content
            )
            plugin = CheckIllegalCharacters(fake_context)

            results = list(plugin.run())

            self.assertEqual(len(results), 3)
            self.assertIsInstance(results[0], LinterError)
            self.assertIsInstance(results[1], LinterError)
            self.assertIsInstance(results[2], LinterWarning)
            self.assertEqual(
                results[0].message,
                f"Found illegal character '|' "
                f'in script_tag(name:"{tag}", '
                'value:"Foo|Bar;Baz=Bad.");',
            )
            self.assertEqual(
                results[1].message,
                f"Found illegal character ';' "
                f'in script_tag(name:"{tag}", '
                'value:"Foo|Bar;Baz=Bad.");',
            )
            self.assertEqual(
                results[2].message,
                f"Found illegal character '=' "
                f'in script_tag(name:"{tag}", '
                'value:"Foo|Bar;Baz=Bad.");',
            )

    def test_fix_ok(self):
        path = Path("some/file.nasl")
        content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo | Bar ; Baz = Test");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )
        fake_context = self.create_file_plugin_context(
            nasl_file=path, file_content=content
        )
        plugin = CheckIllegalCharacters(fake_context)

        results = list(plugin.run())

        fixed_content = (
            '  script_tag(name:"cvss_base", value:"4.0");\n'
            '  script_tag(name:"summary", value:"Foo <pipe> Bar , Baz = '
            'Test");\n'
            '  script_tag(name:"solution_type", value:"VendorFix");\n'
            '  script_tag(name:"solution", value:"meh");\n'
        )

        self.assertEqual(len(results), 3)
        self.assertIsInstance(results[0], LinterError)
        self.assertIsInstance(results[1], LinterError)
        self.assertIsInstance(results[2], LinterWarning)
        self.assertEqual(plugin.new_file_content, fixed_content)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterator

from troubadix.helper.patterns import (
    get_script_tag_patterns,
    get_special_script_tag_patterns,
)
from troubadix.plugin import FilePlugin, LinterError, LinterResult

allowed_dup_dependencies = [
    "GSHB/EL15/GSHB.nasl",
    "gsf/PCIDSS/PCI-DSS.nasl",
    "gsf/PCIDSS/v2.0/PCI-DSS-2.0.nasl",
]


class CheckDuplicatedScriptTags(FilePlugin):
    name = "check_duplicated_script_tags"

    def run(self) -> Iterator[LinterResult]:
        if self.context.nasl_file.suffix == ".inc":
            return

        file_content = self.context.file_content

        if "# troubadix: disable=template_nd_test_files_fps" in file_content:
            return

        special_script_tag_patterns = get_special_script_tag_patterns()

        for tag, pattern in special_script_tag_patterns.items():
            # TBD: script_name might also look like this:
            # script_name("MyVT (Windows)");

            if tag.name == "ADD_PREFERENCE":
                continue

            if tag.name == "DEPENDENCIES":
                file_path = str(self.context.nasl_file)
                if any(f in file_path for f in allowed_dup_dependencies):
                    continue

            match = pattern.finditer(file_content)

            if match:
                # This is allowed, see e.g.
                # gb_netapp_data_ontap_consolidation.nasl
                if tag.value == "dependencies" and "FEED_NAME" in file_content:
                    continue
                if tag.value == "xref":
                    continue

                match = list(match)
                if len(match) > 1:
                    yield LinterError(
                        f"The VT is using the script tag 'script_"
                        f"{tag.value}' multiple number of times.",
                        file=self.context.nasl_file,
                        plugin=self.name,
                    )

        script_tag_patterns = get_script_tag_patterns()
        for tag, pattern in script_tag_patterns.items():
            match = pattern.finditer(file_content)

            if match:
                match = list(match)
                if len(match) > 1:
                    yield LinterError(
                        f"The VT is using the script tag '{tag.value}' "
                        "multiple number of times.",
                        file=self.context.nasl_file,
                        plugin=self.name,
                    )



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase
from unittest.mock import patch

from troubadix.standalone_plugins.changed_cves import (
    compare,
    get_cves_from_content,
)


class ChangedCVEsTest(TestCase):
    def test_get_cves_from_content(self):
        content = (
            "...\n"
            '  script_oid("1.3.6.1.4.1.25623.1.0.705311");\n'
            '  script_version("2023-01-10T10:12:01+0000");\n'
            '  script_cve_id("CVE-2022-32749", "CVE-2022-37392");\n'
            '  script_tag(name:"cvss_base", value:"5.0");\n'
            "..."
        )

        expected_result = {"CVE-2022-32749", "CVE-2022-37392"}

        result = get_cves_from_content(content)

        self.assertEqual(expected_result, result)

    def test_get_cves_from_content_multiline(self):
        content = (
            "...\n"
            '  script_oid("1.3.6.1.4.1.25623.1.0.705311");\n'
            '  script_version("2023-01-10T10:12:01+0000");\n'
            '  script_cve_id("CVE-2022-32749",\n'
            '"CVE-2022-37392");\n'
            '  script_tag(name:"cvss_base", value:"5.0");\n'
            "..."
        )

        expected_result = {"CVE-2022-32749", "CVE-2022-37392"}

        result = get_cves_from_content(content)

        self.assertEqual(expected_result, result)

    def test_get_cves_from_content_empty(self):
        content = (
            "...\n"
            '  script_oid("1.3.6.1.4.1.25623.1.0.705311");\n'
            '  script_version("2023-01-10T10:12:01+0000");\n'
            '  script_tag(name:"cvss_base", value:"5.0");\n'
            "..."
        )

        expected_result = set()

        result = get_cves_from_content(content)

        self.assertEqual(expected_result, result)

    @patch("troubadix.standalone_plugins.changed_cves.get_cves_from_content")
    def test_compare(self, mock):
        mock.side_effect = [
            {"CVE-3333-3333", "CVE-2222-2222", "CVE-1111-1111"},
            {"CVE-1111-1111", "CVE-4444-4444"},
        ]

        expected_missing_cves = ["CVE-2222-2222", "CVE-3333-3333"]
        expected_new_cves = ["CVE-4444-4444"]

        missing_cves, new_cves = compare("", "")

        self.assertEqual(expected_missing_cves, missing_cves)
        self.assertEqual(expected_new_cves, new_cves)



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from enum import Enum
from typing import Dict

from troubadix.helper.helper import SCRIPT_CATEGORIES

# regexp pattern for getting any value of
# script_tag(name:"last_modification", value:"<value>");
LAST_MODIFICATION_ANY_VALUE_PATTERN = (
    r'script_tag\(\s*name\s*:\s*(?P<quote>[\'"])(last_modification)'
    r'(?P=quote)\s*,\s*value\s*:\s*(?P<quote2>[\'"])?'
    r"(?P<value>.*)(?P=quote2)?\s*\)\s*;"
)

# regexp pattern for getting any value of
# script_version("<value>");
SCRIPT_VERSION_ANY_VALUE_PATTERN = (
    r'script_version\(\s*(?P<quote>[\'"])(?P<value>.*)(?P=quote)\s*\);'
)

# regex patterns for script tags
_TAG_PATTERN = (
    r'script_tag\(\s*name\s*:\s*(?P<quote>[\'"])(?P<name>{name})(?P=quote)\s*,'
    r'\s*value\s*:\s*(?P<quote2>[\'"])?(?P<value>{value})(?P=quote2)?\s*\)\s*;'
)


def _get_tag_pattern(
    name: str, *, value: str = r".+?", flags: re.RegexFlag = 0
) -> re.Pattern:
    """
    The returned pattern catches all
    `script_tags(name="{name}", value="{value}");`
    of a specific script tag name

    Arguments:
        name        a SpecialScriptTag Enum type
        value       script tag value (default: at least on char)
        flags       regex flags for compile (default: 0)

    The returned `Match`s by this pattern will have group strings
    .group('name') and .group('value')
    Returns
        `re.Pattern` object
    """
    return re.compile(_TAG_PATTERN.format(name=name, value=value), flags=flags)


class ScriptTag(Enum):
    AFFECTED = "affected"
    CREATION_DATE = "creation_date"
    CVSS_BASE = "cvss_base"
    CVSS_BASE_VECTOR = "cvss_base_vector"
    DEPRECATED = "deprecated"
    IMPACT = "impact"
    INSIGHT = "insight"
    LAST_MODIFICATION = "last_modification"
    QOD = "qod"
    QOD_TYPE = "qod_type"
    SEVERITY_VECTOR = "severity_vector"
    SEVERITY_ORIGIN = "severity_origin"
    SEVERITY_DATE = "severity_date"
    SOLUTION = "solution"
    SOLUTION_TYPE = "solution_type"
    SUMMARY = "summary"
    VULDETECT = "vuldetect"


# pylint: disable=invalid-name
__script_tag_pattern = None

__DATE_VALUE = r"[A-Za-z0-9\:\-\+\,\s\(\)]{44}"

__script_tag_values = {
    ScriptTag.DEPRECATED: r"TRUE",
    ScriptTag.CVSS_BASE_VECTOR: r"AV:[LAN]/AC:[HML]/Au:[NSM]/C:[NPC]/I:"
    r"[NPC]/A:[NPC]",
    ScriptTag.CVSS_BASE: r"(10\.0|[0-9]\.[0-9])",
    ScriptTag.CREATION_DATE: __DATE_VALUE,
    ScriptTag.LAST_MODIFICATION: __DATE_VALUE,
}


def init_script_tag_patterns() -> None:
    # pylint: disable=global-statement
    global __script_tag_pattern

    __script_tag_pattern = dict()

    for tag in ScriptTag:
        flags = 0
        value = __script_tag_values.get(tag)

        if value is None:
            value = r".+?"
            flags = re.MULTILINE | re.DOTALL

        __script_tag_pattern[tag] = _get_tag_pattern(
            name=tag.value, value=value, flags=flags
        )


def get_script_tag_patterns() -> Dict[ScriptTag, re.Pattern]:
    """Returns a dict that contains all precompiled script tag patterns"""
    if not __script_tag_pattern:
        init_script_tag_patterns()

    return __script_tag_pattern


def get_script_tag_pattern(script_tag: ScriptTag) -> re.Pattern:
    """Returns a precompiled script tag pattern by the given tag name

    Arguments:
        script_tag      one of the special script tags from the
                        ScriptTag enum

    The returned `Match`s by this pattern will have group strings
    .group('name') and .group('value')
    Returns
        `re.Pattern` object
    """
    script_tag_patterns = get_script_tag_patterns()
    return script_tag_patterns[script_tag]


_XREF_TAG_PATTERN = (
    r'script_xref\(\s*name\s*:\s*["\'](?P<type>{type})["\']\s*,'
    r'\s*value\s*:\s*["\']?(?P<value>{value})["\']?\s*\)\s*;'
)


def get_xref_pattern(
    name: str,
    *,
    value: str = r".+",
    flags: re.RegexFlag = 0,
) -> re.Pattern:
    """
    The returned pattern catches all
    `script_xref(name="{type}", value="{value}");`

    Arguments:
        name        script xref type e.g. URL
        value       script tag value (default: at least one char)
        flags       regex flags for compile (default: 0)

    The returned `Match`s by this pattern will have group strings
    .group('type') and .group('value')
    Returns
        `re.Pattern` object
    """
    return re.compile(
        _XREF_TAG_PATTERN.format(type=name, value=value),
        flags=flags,
    )


_SPECIAL_TAG_PATTERN = (
    r'script_(?P<name>{name})\s*\((?P<quote99>[\'"])?(?P<value>{value})'
    r"(?P=quote99)?\s*\)\s*;"
)

_XREF_TAG_PATTERN = (
    r'script_xref\(\s*name\s*:\s*(?P<quote>[\'"])(?P<type>{type})'
    r'(?P=quote)\s*,\s*value\s*:\s*(?P<quote2>[\'"])?(?P<value>{'
    r"value})(?P=quote2)?\s*\)\s*;"
)


class SpecialScriptTag(Enum):
    ADD_PREFERENCE = "add_preference"
    BUGTRAQ_ID = "bugtraq_id"
    CATEGORY = "category"
    # script_copyright("Copyright (C) 2021 Greenbone AG");
    COPYRIGHT = "copyright"
    CVE_ID = "cve_id"
    DEPENDENCIES = "dependencies"
    EXCLUDE_KEYS = "exclude_keys"
    FAMILY = "family"  # script_family("FIXME");
    MANDATORY_KEYS = "mandatory_keys"
    NAME = "name"
    OID = "oid"  # script_oid("1.3.6.1.4.1.25623.1.0.XXXXXX");
    REQUIRE_KEYS = "require_keys"
    REQUIRE_PORTS = "require_ports"
    REQUIRE_UDP_PORTS = "require_udp_ports"
    VERSION = "version"  # script_version("YYYY-MM-DDTHH:mm:ss+0000");
    XREF = "xref"


def _get_special_script_tag_pattern(
    name: str, *, value: str = r".+?", flags: re.RegexFlag = 0
) -> re.Pattern:
    """
    The returned pattern catches all `script_<name>("<value>");` tags
    of the given script tag name

    Arguments:
        name        a SpecialScriptTag Enum type
        value       script tag value (default: at least one char)
        flags       regex flags for compile (default: 0)

    The returned `Match`s by this pattern will have group strings
    .group('name') and .group('value')
    Returns
        `re.Pattern` object
    """
    return re.compile(
        _SPECIAL_TAG_PATTERN.format(name=name, value=value), flags=flags
    )


__PORT_VALUE = r"\"(?P<service>[\w\s])+\", (?P<port>\d{1,5})"

# pylint: disable=invalid-name
__special_script_tag_patterns = None

__special_script_tag_values = {
    SpecialScriptTag.OID: r"(?P<oid>([0-9.]+))",
    SpecialScriptTag.CATEGORY: r"(?P<category>("
    rf"{'|'.join([k for k, _ in SCRIPT_CATEGORIES.items()])}))",
    SpecialScriptTag.VERSION: r"[0-9\-\:\+T]{24}|\$Revision: [0-9]+ \$",
    SpecialScriptTag.REQUIRE_PORTS: __PORT_VALUE,
    SpecialScriptTag.REQUIRE_UDP_PORTS: __PORT_VALUE,
    SpecialScriptTag.XREF: r"name:\"(?P<ref_type>[\w\s]+)\","
    r" value:\"(?P<ref>[\w/:= %._\-\?]+)\"",
}


def init_special_script_tag_patterns() -> None:
    # pylint: disable=global-statement
    global __special_script_tag_patterns

    __special_script_tag_patterns = dict()
    for tag in SpecialScriptTag:
        flags = 0
        value = __special_script_tag_values.get(tag)

        if value is None:
            value = r".+?"
            flags = re.MULTILINE

        __special_script_tag_patterns[tag] = _get_special_script_tag_pattern(
            name=tag.value, value=value, flags=flags
        )


def get_special_script_tag_patterns() -> Dict[SpecialScriptTag, re.Pattern]:
    """Returns a dict that contains all precompiled special
    script tag patterns"""
    if not __special_script_tag_patterns:
        init_special_script_tag_patterns()

    return __special_script_tag_patterns


def get_special_script_tag_pattern(
    special_script_tag: SpecialScriptTag,
) -> re.Pattern:
    """Returns a precompiled special script tag pattern by the given tag name

    Arguments:
        special_script_tag      one of the special script tags from the
                                SpecialScriptTag enum

    The returned `Match`s by this pattern will have group strings
    .group('name') and .group('value')
    Returns
        `re.Pattern` object
    """
    special_script_tag_patterns = get_special_script_tag_patterns()
    return special_script_tag_patterns[special_script_tag]


class CommonScriptTagsPattern:
    instance = False

    # re.DOTALL is needed to get multiline tag values
    def __init__(self) -> None:
        self.pattern = _get_tag_pattern(
            name=r"(summary|impact|affected|insight|vuldetect|solution)",
            flags=re.MULTILINE | re.DOTALL,
        )
        self.instance = self


def get_common_tag_patterns() -> re.Pattern:
    if CommonScriptTagsPattern.instance:
        return CommonScriptTagsPattern.instance.pattern
    return CommonScriptTagsPattern().pattern



#  Copyright (c) 2022 Greenbone AG
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterator

from troubadix.helper.patterns import ScriptTag, get_script_tag_pattern
from troubadix.plugin import FilePlugin, LinterError, LinterResult

VALID_QOD_NUM_VALUES = [
    1,
    30,
    50,
    70,
    75,
    80,
    95,
    97,
    98,
    99,
    100,
]

VALID_QOD_TYPES = [
    "exploit",
    "remote_vul",
    "remote_app",
    "package",
    "registry",
    "remote_active",
    "remote_banner",
    "executable_version",
    "remote_analysis",
    "remote_probe",
    "remote_banner_unreliable",
    "executable_version_unreliable",
    "general_note",
    "package_unreliable",
]


class CheckQod(FilePlugin):
    name = "check_qod"

    def run(self) -> Iterator[LinterResult]:
        """
        The script checks the passed VT for the existence / validity of its QoD
        """

        if self.context.nasl_file.suffix == ".inc":
            return

        file_content = self.context.file_content

        if "# troubadix: disable=template_nd_test_files_fps" in file_content:
            return

        qod_pattern = get_script_tag_pattern(ScriptTag.QOD)
        qod_type_pattern = get_script_tag_pattern(ScriptTag.QOD_TYPE)

        match_qod = list(qod_pattern.finditer(file_content))
        match_qod_type = list(qod_type_pattern.finditer(file_content))

        num_matches = len(match_qod) + len(match_qod_type)
        if num_matches < 1:
            yield LinterError(
                "VT is missing QoD or QoD type",
                file=self.context.nasl_file,
                plugin=self.name,
            )
        if num_matches > 1:
            yield LinterError(
                "VT contains multiple QoD values",
                file=self.context.nasl_file,
                plugin=self.name,
            )

        for match in match_qod:
            try:
                qod = int(match.group("value"))
                if qod not in VALID_QOD_NUM_VALUES:
                    yield LinterError(
                        f"{match.group(0)}: '{qod}' is an invalid QoD number"
                        " value. Allowed are"
                        f" {', '.join(str(x) for x in VALID_QOD_NUM_VALUES)}",
                        file=self.context.nasl_file,
                        plugin=self.name,
                    )
            except ValueError:
                yield LinterError(
                    f"{match.group(0)}: '{match.group('value')}' is an invalid"
                    " QoD number value. Allowed are"
                    f" {', '.join(str(x) for x in VALID_QOD_NUM_VALUES)}",
                    file=self.context.nasl_file,
                    plugin=self.name,
                )

        for match in match_qod_type:
            val = match.group("value").replace('"', "")
            if val not in VALID_QOD_TYPES:
                yield LinterError(
                    f"{match.group(0)}: '{match.group('value')}' is an invalid"
                    f" QoD type. Allowed are {', '.join(VALID_QOD_TYPES)}",
                    file=self.context.nasl_file,
                    plugin=self.name,
                )



# Copyright (C) 2021 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Iterator, Optional

from troubadix.helper import CURRENT_ENCODING


@dataclass
class LinterResult:
    """A result found during running a check"""

    message: str
    file: Optional[Path] = None
    plugin: Optional[str] = None
    line: Optional[int] = None


class LinterWarning(LinterResult):
    """A result that is considered a warning"""


class LinterError(LinterResult):
    """A error found during a check"""


class LinterFix(LinterResult):
    """A fix that has been applied"""


class FilePluginContext:
    def __init__(
        self,
        *,
        root: Path,
        nasl_file: Path = None,
    ) -> None:
        self.root = root
        self.nasl_file = nasl_file

        self._file_content = None
        self._lines = None

    @property
    def file_content(self) -> str:
        if not self._file_content:
            self._file_content = self.nasl_file.read_text(
                encoding=CURRENT_ENCODING
            )
        return self._file_content

    @property
    def lines(self) -> Iterable[str]:
        if not self._lines:
            self._lines = self.file_content.splitlines()
        return self._lines


class FilesPluginContext:
    def __init__(self, *, root: Path, nasl_files: Iterable[Path]) -> None:
        self.root = root
        self.nasl_files = nasl_files


class Plugin(ABC):
    """A linter plugin"""

    name: str = None
    description: str = None

    @abstractmethod
    def run(self) -> Iterator[LinterResult]:
        pass

    def fix(self) -> Iterator[LinterResult]:
        return []


class FilesPlugin(Plugin):
    """A plugin that does checks over all files"""

    def __init__(self, context: FilesPluginContext) -> None:
        self.context = context


class FilePlugin(Plugin):
    """A plugin that does checks on single files"""

    def __init__(self, context: FilePluginContext) -> None:
        self.context = context


class FileContentPlugin(FilePlugin):
    """A plugin that does checks on the whole file content"""

    def run(self) -> Iterator[LinterResult]:
        return self.check_content(
            self.context.nasl_file, self.context.file_content
        )

    @abstractmethod
    def check_content(
        self, nasl_file: Path, file_content: str
    ) -> Iterator[LinterResult]:
        pass


class LineContentPlugin(FilePlugin):
    """A plugin that checks file content line by line"""

    def run(self) -> Iterator[LinterResult]:
        return self.check_lines(self.context.nasl_file, self.context.lines)

    @abstractmethod
    def check_lines(
        self,
        nasl_file: Path,
        lines: Iterable[str],
    ) -> Iterator[LinterResult]:
        pass



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase

from troubadix.standalone_plugins.changed_packages.marker import ChangedUpdate
from troubadix.standalone_plugins.changed_packages.package import (
    Direction,
    Package,
    Reasons,
)


class ChangedUpdateTestCase(TestCase):
    def test_mark(self):
        missing_packages = [
            Package("foo", "1.2.3", "DEB11"),
            Package("bar", "4.5.6-deb11u1", "DEB11"),
        ]
        new_packages = [
            Package("foo", "1.2.4", "DEB11"),
            Package("bar", "4.5.6-deb11u2", "DEB11"),
        ]

        expected_missing_packages = [
            Package(
                "foo",
                "1.2.3",
                "DEB11",
                {Reasons.CHANGED_UPDATE: Direction.ACTIVE},
            ),
            Package(
                "bar",
                "4.5.6-deb11u1",
                "DEB11",
                {Reasons.CHANGED_UPDATE: Direction.ACTIVE},
            ),
        ]
        expected_new_packages = [
            Package(
                "foo",
                "1.2.4",
                "DEB11",
                {Reasons.CHANGED_UPDATE: Direction.ACTIVE},
            ),
            Package(
                "bar",
                "4.5.6-deb11u2",
                "DEB11",
                {Reasons.CHANGED_UPDATE: Direction.ACTIVE},
            ),
        ]

        ChangedUpdate.mark(missing_packages, new_packages)

        self.assertEqual(expected_missing_packages, missing_packages)
        self.assertEqual(expected_new_packages, new_packages)

    def test_mark_no_match(self):
        missing_packages = [Package("foo", "1.2.3a", "DEB11")]
        new_packages = [Package("foo", "1.2.3a", "DEB11")]

        expected_missing_packages = [Package("foo", "1.2.3a", "DEB11")]
        expected_new_packages = [Package("foo", "1.2.3a", "DEB11")]

        ChangedUpdate.mark(missing_packages, new_packages)

        self.assertEqual(expected_missing_packages, missing_packages)
        self.assertEqual(expected_new_packages, new_packages)

    def test_mark_no_other_package(self):
        missing_packages = [Package("foo", "1.2.3", "DEB11")]
        new_packages = [Package("foo", "1.2.3-deb11u2", "DEB11")]

        expected_missing_packages = [Package("foo", "1.2.3", "DEB11")]
        expected_new_packages = [Package("foo", "1.2.3-deb11u2", "DEB11")]

        ChangedUpdate.mark(missing_packages, new_packages)

        self.assertEqual(expected_missing_packages, missing_packages)
        self.assertEqual(expected_new_packages, new_packages)



#  Copyright (c) 2022 Greenbone AG
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Iterator

from troubadix.helper.linguistic_exception_handler import (
    PatternCheck,
    TextCheck,
    TextInFileCheck,
    handle_linguistic_checks,
)
from troubadix.plugin import FilePlugin, LinterError, LinterResult

exceptions = [
    PatternCheck(r'(\s+|"|#\s*)[Aa] few (issues|vulnerabilities)'),
    TextCheck("a multiple keyboard "),
    TextCheck("A A S Application Access Server"),
    TextCheck("a Common Vulnerabilities and Exposures"),
    TextCheck("Multiple '/' Vulnerability"),
    TextCheck("an attackers choise"),
    TextCheck("multiple error handling vulnerabilities"),
    # Like seen in 2022/debian/deb_dla_2981.nasl
    TextCheck("a multiple concurrency"),
    # From 2008/debian/deb_1017_1.nasl
    TextCheck("Harald Welte discovered that if a process issues"),
    TextCheck(" a USB Request Block (URB)"),
    # From several Ubuntu LSCs like e.g.:
    # 2021/ubuntu/gb_ubuntu_USN_4711_1.nasl
    TextCheck("An attacker with access to at least one LUN in a multiple"),
    # nb: The regex to catch "this files" might catch this wrongly...
    PatternCheck(r"th(is|ese)\s+filesystem", re.IGNORECASE),
    # Like seen in e.g. 2008/freebsd/freebsd_mod_php4-twig.nasl
    PatternCheck(r'(\s+|")[Aa]\s+multiple\s+of'),
    # WITH can be used like e.g. the following which is valid:
    # "with WITH stack unwinding"
    # see e.g. gb_sles_2021_3215_1.nasl or gb_sles_2021_2320_1.nasl
    PatternCheck(r"with\s+WITH"),
    # Valid sentences
    PatternCheck(
        r"these\s+error\s+(messages|reports|conditions)", re.IGNORECASE
    ),
    PatternCheck(
        r"these\s+file\s+(permissions|overwrites|names|includes)",
        re.IGNORECASE,
    ),
    # nb: Valid sentence
    TextInFileCheck(
        "2012/gb_VMSA-2010-0007.nasl",
        "e. VMware VMnc Codec heap overflow vulnerabilities\n\n"
        "  Vulnerabilities in the",
    ),
    TextInFileCheck("gb_opensuse_2018_1900_1.nasl", "(Note that"),
    # e.g.:
    # multiple cross-site request forgery (CSRF) vulnerabilities
    # Multiple cross-site request forgery vulnerabilities
    PatternCheck(
        r"multiple\s+cross(\s+|-)site(\s+|-)request(\s+|-)forgery",
        re.IGNORECASE,
    ),
    # multiple HTTP request smuggling issues
    # multiple request smuggling vulnerabilities
    PatternCheck(
        r"multiple(\s+HTTP)?\s+request\s+smuggling",
        re.IGNORECASE,
    ),
    # multiple request parameters
    # creates multiple request streams
    PatternCheck(
        r"multiple\s+request\s+(parameters|streams)",
        re.IGNORECASE,
    ),
    # Multiple server-side request forgery (SSRF) vulnerabilities
    # Multiple server-side request forgery vulnerabilities
    # Multiple Server-Side Request Forgery (SSRF) (CVE-2020-4786, CVE-2020-4787)
    PatternCheck(
        r"multiple\s+server(\s+|-)side(\s+|-)request(\s+|-)forgery",
        re.IGNORECASE,
    ),
    # when a page requests a
    # If a report requests external
    # when a client requests a date/time that
    # If a client requests DNS SEC records
    # when a client requests an interactive session
    # a few requests
    PatternCheck(
        r"\s+a\s+(page|client|report|few)\s+requests",
        re.IGNORECASE,
    ),
    # From 2021/mozilla/gb_mozilla_firefox_mfsa_2012-63_lin.nasl
    TextCheck("when an element with a 'requiredFeatures' attribute"),
    # e.g. Multiple Products $something Vulnerability
    PatternCheck(r"Multiple\s+Products.+Vulnerability", re.IGNORECASE),
    # From 2009/ubuntu/gb_ubuntu_USN_592_1.nasl:
    # The default has been changed to prompt the user each time a website
    # requests a client certificate.
    TextCheck("a website requests a client certificate"),
]


def get_grammer_pattern() -> re.Pattern:
    return re.compile(
        r".*("
        r"refer\s+(the\s+)?Reference|"
        r"\s+an?\s+(multiple|errors)|"
        r"the\s+(References?\s+link|multiple\s+flaw)|"
        r"multiple\s+(unknown\s+)?("
        r"vulnerability|flaw|error|problem|issue|feature|request)(\.|\s+)|"
        r"\s+(with\s+with|and\s+and|this\s+this|for\s+for|as\s+as|a\s+a"
        r"|of\s+of|to\s+to|an\s+an|the\s+the|is\s+is|in\s+in|are\s+are|have"
        r"\s+have|has\s+has|that\s+that)\s+|"
        r"vulnerabilit(y|ies)\s+vulnerabilit(y|ies)|"
        r"links\s+mentioned\s+in(\s+the)?\s+reference|"
        r"\s+an?(\s+remote)?(\s+(un)?authenticated)?\s+attackers|"
        # e.g. "this flaws"
        r"this\s+(vulnerabilities|(flaw|error|problem|issue|feature|file|"
        r"request)s)|"
        # e.g. "these flaw "
        r"these\s+(vulnerability|(flaw|error|problem|issue|feature|file|"
        r"request)\s+)|"
        r"\s+or\s+not\.?(\"\);)?$|"
        r"from(\s+the)?(\s+below)?mentioned\s+References?\s+link|"
        r"software\s+it\s+fail|"
        r"references\s+(advisor|link)|"
        r"The\s+multiple\s+(vulnerabilit|flaw|error|problem|issue|feature)|"
        r"(vulnerability|flaw|error|problem|issue|feature)\s+exist\s+|"
        r"(vulnerabilitie|flaw|error|problem|issue|feature)s\s+exists|"
        r"multiple\s+[^\s]+((and\s+)?[^\s]+)?\s+("
        r"vulnerability|flaw|error|problem|issue|feature|request)(\.|\s+)|"
        r"(\s+|^|\"|- )A\s+[^\s]*((and\s+)?[^\s]+\s+)?("
        r"vulnerabilitie|flaw|error|problem|issue|feature|request)s|"
        r"(\s+|^|\"|- )An?\+(unspecified|multiple|unknown)\s+("
        r"vulnerabilitie|flaw|error|problem|issue)s|"
        r"is\s+(prone|vulnerable|affected)\s+(to|by)\s+("
        r"unspecified|XML\s+External\s+Entity|integer\s+(und|ov)erflow|"
        r"DLL\s+hijacking|(hardcoded?|default)\s+credentials?|open[\s-]+"
        r"redirect(ion)?|user\s+enumeration|arbitrary\s+file\s+read|memory"
        r"\s+corruption|use[\s-]+after[\s-]+free|man[\s-]+in[\s-]+the[\s-]"
        r"+middle(\s+attack)?|cross[\s-]+site[\s-]+(scripting(\s+\(XSS\))?"
        r"|request[\s-]+forgery(\s+\(CSRF\))?)|denial[\s-]+of[\s-]+service"
        r"|information\s+disclosure|(path|directory)\s+traversal|"
        r"(arbitrary\s+|remote\s+)?((code|command)\s+(execution|injection)"
        r"|file\s+inclusion)|SQL\s+injection|security|(local )?privilege"
        r"[\s-]+(escalation|elevation)|(authentication|security|access)"
        r"\s+bypass|(buffer|heap)\s+overflow)\s+vulnerability|"
        # e.g.:
        # "is prone a to denial of service (DoS) vulnerability"
        # "is prone an information disclosure vulnerability"
        r"\s+(is|are)\s+(prone|vulnerable|affected)\s+an?\s+|"
        # e.g.:
        # "Sends multiple HTTP request and checks the responses."
        # "Sends multiple HTTP GET request and checks the responses."
        r"multiple\s+([^ ]+\s+)?([^ ]+\s+)?request\s+|"
        # e.g. "is prone to a security bypass vulnerabilities"
        r"is\s+prone\s+to\s+an?\s+[^\s]+\s+([^\s]+\s+)?vulnerabilities" r").*",
        re.IGNORECASE,
    )


class CheckGrammar(FilePlugin):
    name = "check_grammar"

    def run(self) -> Iterator[LinterResult]:
        """This script checks the passed VT / Include for common grammar
        problems

        Args:
            nasl_file:    The VT / Include that is going to be checked
            file_content: The content of the file that is going to be
                          checked
        """
        pattern = get_grammer_pattern()

        for match in pattern.finditer(self.context.file_content):
            if match:
                if handle_linguistic_checks(
                    str(self.context.nasl_file), match.group(0), exceptions
                ):
                    continue

                yield LinterError(
                    "VT/Include has the following grammar problem:"
                    f" {match.group(0)}",
                    file=self.context.nasl_file,
                    plugin=self.name,
                )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Iterator

from troubadix.helper.patterns import (
    _get_special_script_tag_pattern,
    _get_tag_pattern,
    get_xref_pattern,
)
from troubadix.plugin import FileContentPlugin, LinterError, LinterResult

# For the future
SPECIAL_SCRIPT_TAG_LIST = []


class CheckScriptCallsEmptyValues(FileContentPlugin):
    name = "check_script_calls_empty_values"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        """
        Checks for empty 'value:""' in script calls. Excepted from this is
        script_add_preferences().
        """
        if (
            nasl_file.suffix == ".inc"
            or "# troubadix: disable=template_nd_test_files_fps" in file_content
        ):
            return

        matches = _get_tag_pattern(name=r".*", value=r"").finditer(file_content)
        for match in matches:
            yield LinterError(
                f"{match.group(0)} does not contain a value",
                file=nasl_file,
                plugin=self.name,
            )

        matches = get_xref_pattern(name=r".*", value=r"").finditer(file_content)
        for match in matches:
            yield LinterError(
                f"{match.group(0)} does not contain a value",
                file=nasl_file,
                plugin=self.name,
            )

        for call in SPECIAL_SCRIPT_TAG_LIST:
            matches = _get_special_script_tag_pattern(
                name=call.value, value=""
            ).finditer(file_content)
            for match in matches:
                yield LinterError(
                    f"{match.group(0)} does not contain a value",
                    file=nasl_file,
                    plugin=self.name,
                )



#  Copyright (c) 2022 Greenbone AG
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import re
from pathlib import Path
from typing import Iterator

from troubadix.helper import CURRENT_ENCODING
from troubadix.helper.patterns import (
    LAST_MODIFICATION_ANY_VALUE_PATTERN,
    SCRIPT_VERSION_ANY_VALUE_PATTERN,
    ScriptTag,
    SpecialScriptTag,
    get_script_tag_pattern,
    get_special_script_tag_pattern,
)
from troubadix.plugin import (
    FileContentPlugin,
    LinterError,
    LinterFix,
    LinterResult,
)


class CheckScriptVersionAndLastModificationTags(FileContentPlugin):
    name = "check_script_version_and_last_modification_tags"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        """The script checks if the passed VT has a correct syntax of the
        following two tags:

        - script_version();
        - script_tag(name:"last_modification", value:"");

        An error will be thrown if the syntax of those two tags does not match
        the requirements.

        Args:
            nasl_file: The VT that shall be checked
            file_content: The content of the VT that shall be checked
        """
        if (
            nasl_file.suffix == ".inc"
            or "# troubadix: disable=template_nd_test_files_fps" in file_content
        ):
            return

        self.fix_last_modification_and_version = False

        match_script_version_any = re.search(
            pattern=SCRIPT_VERSION_ANY_VALUE_PATTERN,
            string=file_content,
        )
        if not match_script_version_any:
            yield LinterError(
                "VT is missing script_version();.",
                file=self.context.nasl_file,
                plugin=self.name,
            )
            return

        self.old_script_version = match_script_version_any.group(0)
        self.old_script_version_value = match_script_version_any.group("value")

        # script_version("2019-03-21T12:19:01+0000");")
        version_pattern = get_special_script_tag_pattern(
            SpecialScriptTag.VERSION
        )
        version_match = version_pattern.search(file_content)

        if not version_match:
            # check for old format:
            # script_version("$Revision: 12345 $");
            revision_pattern = re.compile(
                r'script_(?P<name>version)\s*\((?P<quote99>[\'"])?(?P<value>\$'
                r"Revision:\s*[0-9]{1,6}\s* \$)(?P=quote99)?\s*\)\s*;"
            )
            revision_match = revision_pattern.search(file_content)
            if not revision_match:
                self.fix_last_modification_and_version = True
                yield LinterError(
                    "VT is using a wrong script_version(); syntax.",
                    file=nasl_file,
                    plugin=self.name,
                )

        match_last_modification_any_value = re.search(
            pattern=LAST_MODIFICATION_ANY_VALUE_PATTERN,
            string=file_content,
        )

        if not match_last_modification_any_value:
            self.fix_last_modification_and_version = False
            yield LinterError(
                'VT is missing script_tag(name:"last_modification".',
                file=self.context.nasl_file,
                plugin=self.name,
            )
            return

        self.old_last_modification = match_last_modification_any_value.group(0)
        self.old_last_modification_value = (
            match_last_modification_any_value.group("value")
        )

        # script_tag(name:"last_modification",
        # value:"2019-03-21 12:19:01 +0000 (Thu, 21 Mar 2019)");
        last_modification_pattern = get_script_tag_pattern(
            ScriptTag.LAST_MODIFICATION
        )
        match_last_modified = last_modification_pattern.search(file_content)

        if not match_last_modified:
            # check for old format:
            # script_tag(name: "last_modification",
            #   value: "$Date: 2021-07-19 12:32:02 +0000 (Mon, 19 Jul 2021) $")
            old_pattern = re.compile(
                r'script_tag\(\s*name\s*:\s*(?P<quote>[\'"])(?P<name>'
                r"last_modification)(?P=quote)\s*,\s*value\s*:\s*(?P<quote2>"
                r'[\'"])?(?P<value>\$Date:\s*[A-Za-z0-9\:\-\+\,\s\(\)]{44}'
                r"\s*\$)(?P=quote2)?\s*\)\s*;"
            )
            match_last_modified = old_pattern.search(file_content)
            if not match_last_modified:
                self.fix_last_modification_and_version = True
                yield LinterError(
                    "VT is is using a wrong syntax for script_tag("
                    'name:"last_modification".',
                    file=nasl_file,
                    plugin=self.name,
                )

    def fix(self) -> Iterator[LinterResult]:
        if not self.fix_last_modification_and_version:
            return

        tag_template = 'script_tag(name:"last_modification", value:"{date}");'
        version_template = 'script_version("{date}");'

        now = datetime.datetime.now(datetime.timezone.utc)

        file_content = self.context.file_content

        # get that version date formatted correctly:
        # "2021-03-24T10:08:26+0000"
        correctly_formatted_version = f"{now:%Y-%m-%dT%H:%M:%S%z}"

        file_content = file_content.replace(
            self.old_script_version,
            version_template.format(date=correctly_formatted_version),
        )

        # get that last modification date formatted correctly:
        # "2021-03-24 10:08:26 +0000 (Wed, 24 Mar 2021)"
        correctly_formatted_last_modification = (
            f"{now:%Y-%m-%d %H:%M:%S %z (%a, %d %b %Y)}"
        )

        file_content = file_content.replace(
            self.old_last_modification,
            tag_template.format(date=correctly_formatted_last_modification),
        )

        self.context.nasl_file.write_text(
            file_content, encoding=CURRENT_ENCODING
        )

        yield LinterFix(
            f"Replaced last_modification {self.old_last_modification_value} "
            f"with {correctly_formatted_last_modification} and script_version "
            f"{self.old_script_version_value} with "
            f"{correctly_formatted_version}.",
            file=self.context.nasl_file,
            plugin=self.name,
        )



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

from troubadix.standalone_plugins.changed_packages.package import (
    Direction,
    Package,
    Reasons,
)

from .marker import Marker


class AddedRelease(Marker):
    @staticmethod
    def mark(old_packages: List[Package], new_packages: List[Package]):
        # Example: ...2015/debian/deb_3257.nasl now has DEB8 next to DEB7
        new_releases = set([package.release for package in new_packages])
        old_releases = set([package.release for package in old_packages])

        for new_release in new_releases.difference(old_releases):
            for package in new_packages:
                if package.release != new_release:
                    continue

                package.reasons[Reasons.ADDED_RELEASE] = Direction.ACTIVE



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import tempfile
import unittest
from pathlib import Path
from typing import Iterable
from unittest.mock import MagicMock

from troubadix.plugin import FilePluginContext, FilesPluginContext


class TemporaryDirectory:
    """A wrapper around tempdir.TemporaryDirectory to return a Path
    when using the with statement
    """

    def __init__(self) -> None:
        self._tempdir = tempfile.TemporaryDirectory()

    def __enter__(self) -> Path:
        return Path(self._tempdir.__enter__())

    def __exit__(self, exc, value, tb) -> None:
        self._tempdir.__exit__(exc, value, tb)

    def __fspath__(self):
        return self._tempdir.name

    def cleanup(self):
        self._tempdir.cleanup()


class PluginTestCase(unittest.TestCase):
    def create_directory(self) -> TemporaryDirectory:
        return TemporaryDirectory()

    def create_file_plugin_context(
        self,
        *,
        nasl_file: Path = None,
        file_content: str = None,
        lines: Iterable[str] = None,
        root: Path = None,
    ) -> FilePluginContext:
        """Create a FilePluginContext mock"""
        fake_context = MagicMock()
        fake_context.nasl_file = nasl_file
        fake_context.file_content = file_content
        fake_context.lines = lines
        fake_context.root = root
        return fake_context

    def create_files_plugin_context(
        self,
        *,
        nasl_files: Iterable[Path] = None,
        root: Path = None,
    ) -> FilesPluginContext:
        """Create a FilePluginContext mock"""
        fake_context = MagicMock()
        fake_context.nasl_files = nasl_files
        fake_context.root = root
        return fake_context



#  Copyright (c) 2022 Greenbone AG
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pathlib import Path

from troubadix.plugin import LinterError
from troubadix.plugins.spelling import CheckSpelling

from . import PluginTestCase


class CheckSpellingTestCase(PluginTestCase):
    def test_ok(self):
        nasl_file = Path(__file__).parent / "test.nasl"
        fake_context = self.create_files_plugin_context(nasl_files=[nasl_file])
        plugin = CheckSpelling(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 0)

    def test_nok(self):
        nasl_file = Path(__file__).parent / "test_files" / "fail_spelling.nasl"
        fake_context = self.create_files_plugin_context(nasl_files=[nasl_file])
        plugin = CheckSpelling(fake_context)

        results = list(plugin.run())

        self.assertEqual(len(results), 3)

        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            f"{nasl_file}:1: soltuion ==> solution",
            results[0].message,
        )

        self.assertIsInstance(results[1], LinterError)
        self.assertEqual(
            f"{nasl_file}:1: aviaalable ==> available",
            results[1].message,
        )

        self.assertIsInstance(results[2], LinterError)
        self.assertEqual(
            f"{nasl_file}:2: upated ==> updated",
            results[2].message,
        )

    def test_local_files_nok(self):
        codespell_additions = Path("codespell.additions")
        codespell_additions.write_text("", encoding="utf-8")

        nasl_file = Path(__file__).parent / "test_files" / "fail_spelling.nasl"
        fake_context = self.create_files_plugin_context(nasl_files=[nasl_file])
        plugin = CheckSpelling(fake_context)

        results = list(plugin.run())

        codespell_additions.unlink()

        self.assertEqual(len(results), 2)
        self.assertIsInstance(results[0], LinterError)
        self.assertEqual(
            f"{nasl_file}:1: soltuion ==> solution",
            results[0].message,
        )
        self.assertEqual(
            f"{nasl_file}:2: upated ==> updated",
            results[1].message,
        )



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess


def git(*args) -> str:
    # git diff output uses raw bytes
    return subprocess.run(
        ["git"] + list(args),
        capture_output=True,
        encoding="latin-1",
        check=True,
    ).stdout


def get_merge_base(*commits: str):
    return git("merge-base", *commits).strip()



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Iterator

from troubadix.helper.patterns import (
    ScriptTag,
    SpecialScriptTag,
    get_script_tag_pattern,
    get_special_script_tag_pattern,
)
from troubadix.plugin import FileContentPlugin, LinterError, LinterResult

MANDATORY_TAGS = [ScriptTag.SUMMARY]
MANDATORY_SPECIAL_TAGS = [
    SpecialScriptTag.NAME,
    SpecialScriptTag.VERSION,
    SpecialScriptTag.CATEGORY,
    SpecialScriptTag.FAMILY,
    SpecialScriptTag.COPYRIGHT,
]


class CheckScriptTagsMandatory(FileContentPlugin):
    name = "check_script_tags_mandatory"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        """
        This script checks for the existence of the following
        mandatory tags:
        - summary
        and special tags:
        - script_name
        - script_version
        - script_category
        - script_family
        - script_copyright
        """
        if (
            nasl_file.suffix == ".inc"
            or "# troubadix: disable=template_nd_test_files_fps" in file_content
        ):
            return

        for tag in MANDATORY_TAGS:
            if not get_script_tag_pattern(tag).search(file_content):
                yield LinterError(
                    "VT does not contain the following mandatory tag: "
                    f"'script_{tag.value}'",
                    file=nasl_file,
                    plugin=self.name,
                )

        for special_tag in MANDATORY_SPECIAL_TAGS:
            if not get_special_script_tag_pattern(special_tag).search(
                file_content
            ):
                yield LinterError(
                    "VT does not contain the following mandatory tag: "
                    f"'script_{special_tag.value}'",
                    file=nasl_file,
                    plugin=self.name,
                )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from troubadix.helper import CURRENT_ENCODING
from troubadix.plugin import LinterError
from troubadix.plugins.encoding import CheckEncoding

from . import PluginTestCase


class CheckEncodingTestCase(PluginTestCase):
    def test_ok(self):
        with self.create_directory() as tempdir:
            path = tempdir / "file.nasl"

            # It seems, that these are the only valid characters for this
            path.write_text(
                "!\"#$%&'()*+,-./0123456789:;<=>"
                "?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`"
                "abcdefghijklmnopqrstuvwxyz{|}~",
                encoding="utf-8",
            )
            content = path.read_text(encoding=CURRENT_ENCODING)
            fake_context = self.create_file_plugin_context(
                nasl_file=path, file_content=content, lines=content.splitlines()
            )
            plugin = CheckEncoding(fake_context)

            results = list(plugin.run())

            self.assertEqual(len(results), 0)

    def test_some_invalid_characters(self):
        with self.create_directory() as tempdir:
            path = tempdir / "file.nasl"

            path.write_text(
                "ȺȺȺȺʉʉʉʉϾϾϾϾ¡¢£¤¥¦§¨©ª«¬­®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄ"
                "ÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ",
                encoding="utf-8",
            )
            content = path.read_text(encoding=CURRENT_ENCODING)
            fake_context = self.create_file_plugin_context(
                nasl_file=path, file_content=content, lines=content.splitlines()
            )
            plugin = CheckEncoding(fake_context)

            results = list(plugin.run())

            self.assertEqual(len(results), 3)

            self.assertIsInstance(results[0], LinterError)
            self.assertEqual(
                "VT uses a wrong encoding. "
                "Allowed encodings are ascii, latin_1.",
                results[0].message,
            )
            self.assertEqual(
                "Found invalid character",
                results[1].message,
            )
            self.assertEqual(
                "Found invalid character",
                results[2].message,
            )

    def test_invisible_whitespace(self):
        with self.create_directory() as tempdir:
            path = tempdir / "file.nasl"

            # It seems, that these are the only valid characters for this
            path.write_text(
                "Local Privilege Escalation Vulnerability",
                encoding="utf-8",
            )
            content = path.read_text(encoding=CURRENT_ENCODING)
            fake_context = self.create_file_plugin_context(
                nasl_file=path, file_content=content, lines=content.splitlines()
            )
            plugin = CheckEncoding(fake_context)

            results = list(plugin.run())

            self.assertEqual(len(results), 1)
            self.assertIsInstance(results[0], LinterError)
            self.assertEqual(
                "VT uses a wrong encoding. "
                "Allowed encodings are ascii, latin_1.",
                results[0].message,
            )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from pathlib import Path
from typing import Iterator

from troubadix.plugin import (
    FileContentPlugin,
    LinterError,
    LinterResult,
    LinterWarning,
)


class CheckUsingDisplay(FileContentPlugin):
    name = "check_using_display"

    def check_content(
        self,
        nasl_file: Path,
        file_content: str,
    ) -> Iterator[LinterResult]:
        if "# troubadix: disable=template_nd_test_files_fps" in file_content:
            return

        display_matches = re.finditer(
            r".*(display\s*\([^)]+\)\s*;)", file_content
        )
        if display_matches is None:
            return

        for display_match in display_matches:
            if display_match is not None and display_match.group(0):
                dis_match = display_match.group(0)
                file_name = str(nasl_file)

                # Known false positives because the if_match check above can't
                # detect something like e.g.:
                # if( debug )
                #   display("foo");
                if (
                    "ssh_func.inc" in file_name
                    and "display( debug_str )" in dis_match
                ):
                    continue

                if (
                    "gb_treck_ip_stack_detect.nasl" in file_name
                    and 'display("---[' in dis_match
                ):
                    continue

                if (
                    "ike_isakmp_func.inc" in file_name
                    and 'display( "---[' in dis_match
                ):
                    continue

                if (
                    "pcap_func.inc" in file_name
                    and 'display( "---[' in dis_match
                ):
                    continue

                if (
                    "os_eol.inc" in file_name
                    and 'display( "DEBUG: Base CPE' in dis_match
                ):
                    continue

                if (
                    "gsf/dicom.inc" in file_name
                    or "enterprise/dicom.inc" in file_name
                    or "global_settings.inc" in file_name
                    or "rdp.inc" in file_name
                    or "bin.inc" in file_name
                ):
                    continue

                if (
                    "DDI_Directory_Scanner.nasl" in file_name
                    and ":: Got a" in dis_match
                ):
                    continue

                if_comment_match = re.search(
                    r"(if[\s]*\(|#).*display\s*\(", dis_match
                )
                if (
                    if_comment_match is not None
                    and if_comment_match.group(0) is not None
                ):
                    yield LinterWarning(
                        f"VT is using a display() function which "
                        f"is protected by a comment or an if statement at: "
                        f"{dis_match}.",
                        file=nasl_file,
                        plugin=self.name,
                    )
                else:
                    yield LinterError(
                        f"VT/Include is using a display() "
                        f"function at: {dis_match}",
                        file=nasl_file,
                        plugin=self.name,
                    )



# Copyright (C) 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from argparse import ArgumentError
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict


class Direction(Enum):
    ACTIVE = 1
    PASSIVE = 2


class Reasons(str, Enum):
    DROPPED_ARCHITECTURE = "Dropped architecture"
    ADDED_EPOCH = "Added epoch"
    CHANGED_UPDATE = "Changed update"
    ADDED_UDEB = "Added udeb package"
    ADDED_RELEASE = "Added a new release"

    def __str__(self) -> str:
        return self.name.lower().replace("_", "-")

    @classmethod
    def from_cli_argument(cls, cli_argument: str):
        try:
            return cls[cli_argument.upper().replace("-", "_")]
        except KeyError as error:
            raise ArgumentError(
                None, f"Invalid reason '{cli_argument}'"
            ) from error


@dataclass()
class Package:
    name: str
    version: str
    release: str
    reasons: Dict[Reasons, Direction] = field(default_factory=dict)

    def __hash__(self) -> int:
        return hash((self.name, self.version, self.release))

    def __eq__(self, other: "Package") -> bool:
        return (
            self.name == other.name
            and self.version == other.version
            and self.release == other.release
            and self.reasons == other.reasons
        )

    def __lt__(self, other: "Package") -> bool:
        # Sort by release first, then the other fields
        if self.release != other.release:
            return self.release < other.release
        if self.name != other.name:
            return self.name < other.name
        if self.version != other.version:
            return self.version < other.version

        return False

    def __str__(self) -> str:
        result = f"{self.name : <50} {self.version : <40} {self.release : <10}"

        reasons = ", ".join(
            f"{change}"
            f"{' in new package' if direction == Direction.PASSIVE else ''}"
            for change, direction in self.reasons.items()
        )
        result += f"{reasons : <10}"

        return result



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Iterator, List

from troubadix.helper import CURRENT_ENCODING
from troubadix.helper.patterns import get_common_tag_patterns
from troubadix.plugin import (
    FilePlugin,
    LinterError,
    LinterFix,
    LinterResult,
    LinterWarning,
)

# import magic


# ;                 can not be displayed in GSA, within
#                   (summary|impact|affected|insight|vuldetect|solution)
# | =               are delimiter in the internal VT cache
#                   (e.g. Tag1=Foo|Tag2=Bar|Tag3=Baz)
#                   in all script_tag(name:"", value:"")
FORBIDDEN_CHARS = {
    "|": (LinterError, "<pipe>"),
    ";": (LinterError, ","),
    "=": (LinterWarning, None),
}


def check_forbidden(match: re.match) -> List[str]:
    """Check the given tag for forbidden characters

    Args:
        match (re.match): The tag to check

    Returns:
        List[str]: The list of forbidden characters that were found
    """
    return [char for char in FORBIDDEN_CHARS if char in match.group("value")]


def fix_forbidden(
    match: re.Match, found_forbidden_characters: List[str]
) -> str:
    """Returns the fixed version of the tag with
       the forbidden characters replaced

    Args:
        match (re.Match): The tag containing forbidden characters
        found_forbidden_characters (List[str]): The list of forbidden
                                                characters found

    Returns:
        str: The sanitized tag
    """
    line: str = match.group(0)
    value: str = match.group("value")
    for char in found_forbidden_characters:
        _, replacement = FORBIDDEN_CHARS[char]
        if replacement:
            value = value.replace(char, replacement)
    return line.replace(match.group("value"), value)


class CheckIllegalCharacters(FilePlugin):
    name = "check_illegal_characters"

    def run(self) -> Iterator[LinterResult]:
        """
        Currently the following chars are not allowed in
        every script_tag(name:"", value:"") :
        """

        if self.context.nasl_file.suffix == ".inc":
            return

        pattern = get_common_tag_patterns()

        self.new_file_content = None

        tag_matches = pattern.finditer(self.context.file_content)
        if tag_matches:
            for match in tag_matches:
                if match and match.group(0) is not None:
                    found_forbidden_characters = check_forbidden(match)
                    if found_forbidden_characters:
                        self.new_file_content = (
                            self.context.file_content.replace(
                                match.group(0),
                                fix_forbidden(
                                    match, found_forbidden_characters
                                ),
                            )
                        )

                        for forbidden_char in found_forbidden_characters:
                            result, _ = FORBIDDEN_CHARS[forbidden_char]
                            yield result(
                                f"Found illegal character '{forbidden_char}' "
                                f"in {match.group(0)}",
                                file=self.context.nasl_file,
                                plugin=self.name,
                            )

    def fix(self) -> Iterator[LinterResult]:
        if not self.new_file_content:
            return

        self.context.nasl_file.write_text(
            self.new_file_content, encoding=CURRENT_ENCODING
        )

        yield LinterFix(
            "Replaced Illegal Characters.",
            file=self.context.nasl_file,
            plugin=self.name,
        )



#  Copyright (c) 2022 Greenbone AG
#
#  SPDX-License-Identifier: GPL-3.0-or-later
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import re
from contextlib import redirect_stdout
from pathlib import Path
from typing import Iterator, Tuple

from codespell_lib import main as codespell_main

from troubadix.helper.linguistic_exception_handler import (
    PatternInFileCheck,
    PatternInFilePatternCheck,
    PatternInFilesCheck,
    PatternsCheck,
    PatternsInFileCheck,
    PatternsInFilePatternCheck,
    handle_linguistic_checks,
)
from troubadix.plugin import FilesPlugin, LinterError, LinterResult

plugin_path = Path(__file__).parent.resolve()
codespell_config_path = (plugin_path.parent / "codespell").resolve()


exceptions = [
    # From /Policy which is just a huge blob of text
    # and too large for codespell.exclude:
    PatternsInFileCheck(
        "policy_file_checksums_win.nasl",
        [r"nD\s+==>\s+and, 2nd", r"oD\s+==>\s+of"],
    ),
    # Same for a few other files:
    PatternInFileCheck("smtp_AV_42zip_DoS.nasl", r"BA\s+==>\s+BY, BE"),
    PatternInFileCheck("bad_ssh_host_keys.inc", r"ba\s+==>\s+by, be"),
    PatternsInFileCheck(
        "wmi_misc.inc", [r"BA\s+==>\s+BY, BE", r"OD\s+==>\s+OF"]
    ),
    PatternInFilesCheck(
        ["ssl_funcs.inc", "gb_ssl_tls_cert_details.nasl"],
        r"fpr\s+==>\s+for, far, fps",
    ),
    # Codespell has currently cna->can in the dictionary.txt
    # which is causing false positives for CNA (widely used term
    # in VTs) because codespell doesn't look at the casing. For
    # now we're excluding any uppercase "CNA" results because
    # these are usually false positives we don't want to report.
    # Similar with "RO" because that is also very likely a false
    # positive.
    PatternsCheck([r"CNA\s+==>\s+CAN", r"RO\s+==>\s+TO"]),
    # Name of a Huawei product
    PatternInFilesCheck(
        ["gb_huawei", "telnetserver_detect_type_nd_version.nasl"],
        r"eSpace\s+==>\s+escape",
        re.IGNORECASE,
    ),
    # "ure" is a Debian package, again too many hits for
    # codespell.exclude.
    PatternInFilePatternCheck(
        r"(deb_(dla_)?[0-9]+(_[0-9]+)?|gb_ubuntu_.+)\.nasl",
        r"ure\s+==>\s+sure",
    ),
    # gsf/PCIDSS VTs are currently using some german text parts
    # nb: codespell seems to have some issues with
    # german umlauts in the codespell.exclude so a few of these
    # were also excluded here instead of directly
    # via codespell.exclude.
    PatternInFilesCheck(
        [
            "PCIDSS/",
            "GSHB/",
            "attic/PCIDSS_",
            "ITG_Kompendium/",
            "Policy/ITG/",
        ],
        r"(sie|ist|oder|prozess|manuell|unter|funktion|"
        r"alle|als|tage|lokale|uptodate|paket|titel|ba|"
        r"ordner|modul|interaktive|programm|explizit|"
        r"normale|applikation|attributen|lokal|signatur|"
        r"modell|klick|generell)\s+==>\s+",
        re.IGNORECASE,
    ),
    # False positives in the gsf/PCIDSS and GSHB/ VTs:
    # string('\nIn the file sent\nin milliseconds
    # There are too many hits to maintain
    # them in codespell.exclude so exclude them for now here.
    PatternInFilesCheck(
        ["PCIDSS/", "GSHB/", "attic/PCIDSS_", "Policy/ITG/"],
        r"n[iI]n\s+==>\s+inn",
    ),
    # False positive in this VT in German example responses.
    PatternInFileCheck(
        "gb_exchange_server_CVE-2021-26855_active.nasl", r"ist\s+==>\s+is"
    ),
    # Mostly a false positive in LSCs because of things like
    # "ALSA: hda" or a codec called "Conexant". There are too
    # many hits to maintain them in codespell.exclude so exclude
    # them for now here.
    PatternInFilePatternCheck(
        r"gb_(sles|(open)?suse|ubuntu_USN)_.+\.nasl",
        r"(hda|conexant)\s+==>\s+(had|connexant)",
        text_pattern_flags=re.IGNORECASE,
    ),
    # Jodie Chancel is a security researcher who is mentioned
    # many times in Mozilla advisories
    PatternInFilePatternCheck(
        r"gb_mozilla_firefox_mfsa_\d{4}-\d{2,4}_lin\.nasl",
        r"Chancel\s+==>\s+Cancel",
    ),
    # Look like correct as this is also in dictionary_rare.txt
    PatternInFileCheck("deb_dla_2896.nasl", r"dependant\s+==>\s+dependent"),
    # Similar to the one above for e.g. SLES.
    # Also exclude "tre", because it's a package name.
    PatternInFilePatternCheck(
        r"mgasa-\d{4}-\d{4}\.nasl",
        r"(hda|tre|conexant)\s+==>\s+(had|tree|connexant)",
        file_pattern_flags=re.IGNORECASE,
    ),
    # Similar to the corrections above, with some additional
    # exclusions like e.g. names
    PatternsInFilePatternCheck(
        r"ELSA-\d{4}-\d{4,5}\.nasl",
        [
            (r"Stange\s+==>\s+Strange", 0),
            (r"chang\s+==>\s+change, charge", 0),
            (r"IST\s+==>\s+IS, IT, ITS, IT'S, SIT, LIST", 0),
            (r"hda\s+==>\s+had", 0),
            (r"Readded\s+==>\s+Read", 0),
            (r"ACI\s+==>\s+ACPI", re.IGNORECASE),
            (r"UE\s+==>\s+USE, DUE", 0),
        ],
    ),
    # NAM / nam is the abbreviation of these products. In
    # netop_infopublic.nasl there is a "nam" parameter.
    PatternInFilePatternCheck(
        r"gb_((cisco|solarwinds)_nam|netiq_access_manager)_",
        r"nam\s+==>\s+name",
        text_pattern_flags=re.IGNORECASE,
    ),
    PatternInFileCheck(
        "/netop_infopublic.nasl", r"nam\s+==>\s+name", flags=re.IGNORECASE
    ),
    # Product names used in a few VTs (no re.IGNORECASE is expected)
    PatternsCheck([r"renderD\s+==>\s+rendered", r"VertX\s+==>\s+vertex"]),
    PatternInFileCheck("_vertx_", r"vertx\s+==>\s+vertex"),
    # This is a class name of Tomcat which is getting checked / mentioned in
    # these, e.g.:
    # org.apache.juli.AsyncFileHandler.directory
    # JULI logging
    PatternInFileCheck("/Tomcat/tomcat_", r"juli\s+==>\s+july"),
    PatternInFileCheck("tomcat", r"JULI\s+==>\s+JULY"),
    # Some valid abbreviation for e.g. Cisco or VLC VTs
    PatternInFileCheck("caf", r"CAF\s+==>\s+CALF"),
]


class CheckSpelling(FilesPlugin):
    name = "check_spelling"

    def _parse_codespell_line(self, line: str) -> Tuple[str, str]:
        if not "==>" in line:
            raise ValueError("Invalid codespell line")

        file, _, correction = line.split(":")

        return file, correction.strip()

    def run(self) -> Iterator[LinterResult]:
        """This script checks, via the codespell library, wether
        the provided nasl files contain spelling errors.
        Certain errors are ignored based on listed exceptions

        Yields:
            Iterator[LinterResult]: The detected spelling errors
        """

        # Overwrite with local repository files if exist
        if Path("codespell.additions").exists():
            codespell_additions = Path("codespell.additions").resolve()
        else:
            codespell_additions = f"{codespell_config_path}/codespell.additions"
        if Path("codespell.exclude").exists():
            codespell_exclude = Path("codespell.exclude").resolve()
        else:
            codespell_exclude = f"{codespell_config_path}/codespell.exclude"
        if Path("codespell.ignore").exists():
            codespell_ignore = Path("codespell.ignore").resolve()
        else:
            codespell_ignore = f"{codespell_config_path}/codespell.ignore"

        batch_size = 10_000
        for i in range(0, len(self.context.nasl_files), batch_size):
            files_parameters = [
                str(nasl_file)
                for nasl_file in self.context.nasl_files[i : i + batch_size]
            ]
            codespell_arguments = [
                "--hard-encoding-detection",
                "--dictionary=-",
                f"--dictionary={codespell_additions}",
                f"--exclude-file={codespell_exclude}",
                f"--ignore-words={codespell_ignore}",
                "--disable-colors",
                "--uri-ignore-words-list=*",
            ] + files_parameters

            with redirect_stdout(io.StringIO()) as codespell_stream:
                codespell_main(*codespell_arguments)

            codespell_output = codespell_stream.getvalue()

            if "Traceback (most recent call last):" in codespell_output:
                yield LinterError(
                    codespell_output,
                    plugin=self.name,
                )

                continue

            codespell_entries = [
                line
                for line in codespell_output.splitlines()
                if not handle_linguistic_checks(
                    *self._parse_codespell_line(line), exceptions
                )
            ]

            for codespell_entry in codespell_entries:
                if "==>" in codespell_entry:
                    yield LinterError(
                        codespell_entry,
                        file=codespell_entry.split(":")[0],
                        plugin=self.name,
                    )



# Copyright (C) 2022 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=fixme

import re
from pathlib import Path
from typing import Iterator

from troubadix.helper.helper import FEED_VERSIONS, is_enterprise_folder
from troubadix.helper.patterns import _get_special_script_tag_pattern
from troubadix.plugin import (
    FilePlugin,
    LinterError,
    LinterResult,
    LinterWarning,
)


class CheckDependencies(FilePlugin):
    name = "check_dependencies"

    def run(
        self,
    ) -> Iterator[LinterResult]:
        """This script checks whether the files used in script_dependencies()
        exist on the local filesystem.
        An error will be thrown if a dependency could not be found.
        """

        if self.context.nasl_file.suffix == ".inc":
            return

        file_content = self.context.file_content

        if "# troubadix: disable=template_nd_test_files_fps" in file_content:
            return

        dependencies_pattern = _get_special_script_tag_pattern(
            "dependencies", flags=re.DOTALL | re.MULTILINE
        )

        root = self.context.root

        matches = dependencies_pattern.finditer(file_content)

        for match in matches:
            if match:
                # Remove single and/or double quotes, spaces
                # and create a list by using the comma as a separator
                # additionally, check and filter for inline comments
                dependencies = []

                for line in match.group("value").splitlines():
                    subject = line[: line.index("#")] if "#" in line else line
                    _dependencies = re.sub(r'[\'"\s]', "", subject).split(",")
                    dependencies += [dep for dep in _dependencies if dep != ""]

                for dep in dependencies:
                    # TODO: gsf/PCIDSS/PCI-DSS.nasl,
                    # gsf/PCIDSS/v2.0/PCI-DSS-2.0.nasl
                    # and GSHB/EL15/GSHB.nasl
                    # are using a variable which we currently
                    # can't handle.
                    if "+d+.nasl" in dep:
                        continue

                    if not any(
                        (root / vers / dep).exists() for vers in FEED_VERSIONS
                    ):
                        yield LinterError(
                            f"The script dependency {dep} could not "
                            "be found within the VTs.",
                            file=self.context.nasl_file,
                            plugin=self.name,
                        )
                        continue

                    dependency = Path(dep)
                    parts = dependency.parts

                    if is_enterprise_folder(parts[0]):
                        # strip parent enterprise folder
                        parts = parts[1:]

                    if len(parts) < 2:
                        # only the filename is contained in parts means
                        # no parent directory
                        continue

                    parent_folder = parts[0]
                    if parent_folder in ["PCIDSS", "Policy", "GSHB"]:
                        yield LinterWarning(
                            f"The script dependency {dep} is in a "
                            "subdirectory, which might be misplaced.",
                            file=self.context.nasl_file,
                            plugin=self.name,
                        )
                    else:
                        yield LinterError(
                            f"The script dependency {dep} is within "
                            "a subdirectory, which is not allowed.",
                            file=self.context.nasl_file,
                            plugin=self.name,
                        )
